# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
from typing import Callable
from kernax.types import Array

def IMQ(x: Array, y: Array, lengthscale: float):
    """Inverse multi-quadratric kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    lengthscale
        Scalar lengthscale / bandwidth

    Returns
    -------

    Value of the kernel function k(x,y)
    """
    return 1.0/jnp.sqrt(1.0 + jnp.sum((x-y)**2)/lengthscale**2)

def Gaussian(x: Array, y: Array, lengthscale: float):
    """Gaussian kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    lengthscale
        Scalar lengthscale / bandwidth

    Returns
    -------

    Value of the kernel function k(x,y)
    """
    return jnp.exp(-0.5*jnp.sum((x-y)**2)/lengthscale**2)

def Energy(x: Array, y: Array):
    """Distance induced kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`

    Returns
    -------

    Value of the kernel function k(x,y)
    """
    return jnp.sqrt(jnp.sum(x**2)) + jnp.sqrt(jnp.sum(y**2)) - jnp.sqrt(jnp.sum((x-y)**2))

def SteinIMQ(x: Array, sx: Array, y: Array, sy: Array, lengthscale: float):
    """Langevin Stein kernel with the IMQ as the underlying kernel

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    sx
        Score functon evaluated at x, vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    sy
        Score function evaluetaed at y, vector of dimension :math:`d`

    Returns
    -------

    Value the Stein IMQ kernel :math:`k_p(x,y)`
    """

    d = len(x)
    sqdist = jnp.sum((x-y)**2)
    qf = 1.0/(1.0 + sqdist/lengthscale**2)
    t3 = jnp.dot(sx, sy) * jnp.sqrt(qf)
    t2 = (1.0/lengthscale**2) * (d + jnp.dot(sx-sy,x-y)) * qf**(3/2)
    t1 = (-3.0/lengthscale**4)*sqdist * qf**(5/2)
    return t1 + t2 + t3

def SteinGaussian(x: Array, sx: Array, y: Array, sy: Array, lengthscale: float):
    """Langevin Stein kernel with the Gaussian kernel as the underlying kernel

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    sx
        Score functon evaluated at x, vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    sy
        Score function evaluetaed at y, vector of dimension :math:`d`

    Returns
    -------

    Value the Stein Gaussian kernel :math:`k_p(x,y)`
    """

    d = len(x)
    sqdist = jnp.sum((x-y)**2)
    kxy = jnp.exp(-0.5*sqdist/lengthscale**2)
    t1 = d/lengthscale**2 - sqdist/lengthscale**4
    t2 = jnp.dot(sx-sy,x-y)/lengthscale**2
    t3 = jnp.dot(sx,sy)
    return kxy*(t1 + t2 + t3)

def GetSteinFn(kernel_fn: Callable):
    """Helper that builds the Stein kernel function :math:`k_p(x,y)`given an arbitrary underlying kernel :math:`k(x,y)`.

    Parameters
    ----------

    kernel_fn
        Callable kernel function of the form :math:`(x,y) \mapsto k(x,y)`. Any hyparameters such as the lengthscale should be fixed with `jax.tree_util.Partial`.

    Returns
    -------

    Callable function of the form :math:`(x, s_x, y, s_y) \mapsto k_p(x, s_x, y, s_y)` where :math:`x` and :math:`y` are the vectors of observations and :math:`s_x` and :math:`s_y` are the associated scores.
    """

    grad1_fn = jax.grad(kernel_fn, argnums=0)
    grad2_fn = jax.grad(kernel_fn, argnums=1)
    hessian_fn = jax.jacobian(grad2_fn, argnums=0)
    def kp_fn(x, sx, y, sy):
        return jnp.dot(sx,sy)*kernel_fn(x,y) + jnp.dot(sx, grad2_fn(x,y)) + jnp.dot(sy, grad1_fn(x,y)) + jnp.trace(hessian_fn(x,y))
    return kp_fn


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
from kernax.kernels import Energy, IMQ, GetSteinFn
from kernax.utils import median_heuristic
from typing import Callable
from kernax.types import Array

def MMD(x: Array, y: Array) -> float:
    """Implements the V-estimator the maximum mean discrepancy.

    Maximum mean discrepancy between two probability distributions :math:`P` and :math:`Q` is defined as:

    .. math::

        \mathrm{MMD}^2(P,Q) = \mathbb{E}_{x\sim P}\mathbb{E}_{y\sim P}k(x,y) + \mathbb{E}_{x\sim Q}\mathbb{E}_{y\sim Q}k(x,y) - 2 \mathbb{E}_{x\sim P}\mathbb{E}_{y\sim Q}k(x,y)

    Given two samples :math:`\{x_i\}_{i=1}^{n}` and :math:`\{y_i\}_{i=1}^{m}`, this function estimates MMD with a V-statistics.

    Examples
    --------

    Assume that we have two samples from Gaussian distributions.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        rng_key1, rng_key2 = jax.random.split(key)

        x = jax.random.normal(rng_key1, (10_000,2))
        y = jax.random.normal(rng_key2, (10_000,2)) + jnp.array([0.1, 0.5])[:,None]


    The maximum mean discrepancy estimated by a V-statistics can be obtained as follows.

    .. code::

        mmd = MMD(x, y)

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    y
        Sample matrix of size :math:`(n, d)`

    Returns
    -------

    The V-estimator of the maximum mean discrepancy between the provided samples.
    """

    kxx = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(x))(x)
    kyy = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(y))(y)
    kxy = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(x))(y)
    mmd = (1.0/x.shape[0]**2)*kxx.sum() + (1.0/y.shape[0]**2)*kyy.sum() - (2.0/(x.shape[0]*y.shape[0]))*kxy.sum()
    return jnp.sqrt(mmd)

def KSD(x: Array, sx: Array, kernel_fn: Callable=None) -> float:
    """Implements the V-estimator of kernelized Stein discrepancy.

    Kernelized Stein discrepancy between two probability distributions is defined as:

    .. math::
        \mathrm{KSD}^2(P,Q) = E_{x\sim Q}E_{y\sim Q} k_p(x,y)

    where :math:`k_p` denotes the Stein kernel.

    Examples
    --------

    Assume that we have a sample from a Gaussian distribution.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (10_000,2)) + jnp.array([1.0, 1.0])[None,:]

    The target distrubution is also a Gaussian distribution but centered. Hence, the scores are obtained as follows.

    .. code::

        score_fn = lambda x: -x
        sx = score_fn(x)

    We can now compute KSD.

    .. code::

        ksd = KSD(x, sx)

    Note that we can also specify the underlying kernel involved in the Stein kernel.

    This function uses the inverse multi-quadratric kernel by default but any other kernel function can be passed as an argument.

    .. code::

        from kernax.kernels import Gaussian

        ksd = KSD(x, sx, Gaussian)

    The lengthscale (or bandwidth) of the underyling `kernel_fn` is automatically set to the median heuristic.

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    sx
        Matrix of size :math:`(n, d)` containing the scores of the sample :math:`x`.
    kernel_fn
        Callable corresponding the the underlying kernel function in the Stein kernel (Default: IMQ kernel).

    Returns
    -------

    The V-estimator of kernelized Stein discrepancy given the provided sample and chosen score function.
    """

    if kernel_fn is None:
        lengthscale = jnp.array(median_heuristic(x))
        kernel_fn = jax.tree_util.Partial(IMQ, lengthscale=lengthscale)

    kp_fn = GetSteinFn(kernel_fn)
    kp = jax.vmap(lambda a, b: jax.vmap(lambda c, d: kp_fn(a,b,c,d))(x,sx))(x,sx)
    ksd = jnp.sqrt(jnp.sum(kp))/x.shape[0]
    return ksd


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import matplotlib.pyplot as plt
try:
    plt.rcParams.update({
            "text.usetex": True,
            "font.family": "sans-serif",
            "font.sans-serif": ["Helvetica"]})
    plt.style.use('seaborn-v0_8')
    FONTSIZE=24
except:
    FONTSIZE=20

import jax
import jax.numpy as jnp
import numpy as np

import kernax
from kernax.utils import median_heuristic
from kernax.toy_mixtures import GaussianMixture

def main(prefix: str, n: int, m: int, d: int, pi: jnp.ndarray, mu: jnp.ndarray, covP: jnp.ndarray, covQ: jnp.ndarray):

    dist_p = GaussianMixture(d=d, pi=pi, mu=mu, cov=covP)
    dist_q = GaussianMixture(d=d, pi=pi, mu=mu, cov=covQ)

    logprob_fn = dist_p.logprob_fn
    score_fn = jax.grad(logprob_fn)

    np.random.seed(1)
    xQ = dist_q.rand(n)
    xP = dist_p.rand(n)
    lengthscale = jnp.array(median_heuristic(xP))

    log_p = jax.vmap(logprob_fn, 0)(xQ)
    score_p = jax.vmap(score_fn, 0)(xQ)
    laplace_log_p = kernax.laplace_log_p_softplus(xQ, logprob_fn)

    st_fn = kernax.SteinThinning(xQ, score_p, lengthscale)
    regst_fn = kernax.RegularizedSteinThinning(xQ, log_p, score_p, laplace_log_p, lengthscale)

    idx_st = st_fn(m)
    idx_regst = regst_fn(m)

    xmax = np.max([np.abs(xQ[:,0].min()), xQ[:,0].max()])
    xmin = -xmax
    ymax = np.max([np.abs(xQ[:,1].min()), xQ[:,1].max()])
    ymin = -ymax

    fig, ax = plt.subplots(1, 1, figsize=(6,6))
    ax.plot(xQ[:,0], xQ[:,1], linestyle="", marker="o", color="k")
    ax.plot(xQ[idx_st,0], xQ[idx_st,1], linestyle="", marker="o", color="r")
    ax.tick_params(labelsize=FONTSIZE)
    ax.set_xlabel(r"$x^{(1)}$", fontsize=FONTSIZE)
    ax.set_ylabel(r"$x^{(2)}$", fontsize=FONTSIZE)
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([xmin, xmax])
    ax.set_title(r"$\mathrm{Stein}$ $\mathrm{Thinning}$", fontsize=FONTSIZE)
    fig.savefig(f"{prefix}_st_n{n}_m{m}_d{d}.png", format="png", bbox_inches="tight")

    fig, ax = plt.subplots(1, 1, figsize=(6,6))
    ax.plot(xQ[:,0], xQ[:,1], linestyle="", marker="o", color="k")
    ax.plot(xQ[idx_regst,0], xQ[idx_regst,1], linestyle="", marker="o", color="r")
    ax.tick_params(labelsize=FONTSIZE)
    ax.set_xlabel(r"$x^{(1)}$", fontsize=FONTSIZE)
    ax.set_ylabel(r"$x^{(2)}$", fontsize=FONTSIZE)
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([xmin, xmax])
    ax.set_title(r"$\mathrm{Regularized}$ $\mathrm{Stein}$ $\mathrm{Thinning}$", fontsize=FONTSIZE)
    fig.savefig(f"{prefix}_rst_n{n}_m{m}_d{d}.png", format="png", bbox_inches="tight")

    # Laplacian correction heatmap
    jac_fn = jax.jacfwd(jax.jacrev(logprob_fn))

    xx = np.linspace(xmin, xmax, 400)
    yy = np.linspace(ymin, ymax, 400)
    XX, YY = np.meshgrid(xx, yy)
    XQ = np.stack([XX.ravel(), YY.ravel()], axis = 1)

    laplacian_correction = jnp.nan_to_num(jax.vmap(jnp.trace)(jnp.clip(jax.vmap(jac_fn)(XQ),a_min=0)))

    fig, ax = plt.subplots(1, 1, figsize=(6,6))
    ax.contourf(XX, YY, np.reshape(laplacian_correction, (400, 400)), cmap = "jet", levels = 30)
    ax.set_title(r"$\mathrm{Laplacian}$ $\mathrm{correction}$ $\mathrm{heatmap}$", fontsize=16)
    fig.savefig(f"{prefix}_laplacian_correction.png", format = "png")

    return xQ, xP, idx_st, idx_regst

if __name__ == "__main__":

    # Figures pathology I
    d = 2
    pi = jnp.array([0.2, 0.8])
    mu = jnp.array([[-3.0]+[0.0]*(d-1), [3.0]+[0.0]*(d-1)])
    covP = [jnp.eye(d)]*2
    covQ = covP

    xQ, xP, idx_st, idx_regst = main(prefix="mog2_unbalanced", n=3000, m=300, d=d, pi=pi, mu=mu, covP=covP, covQ=covQ)

    # Figures pathology II
    d = 2
    pi = jnp.array([0.5]*2)
    mu = jnp.array([[-2.0]+[0.0]*(d-1), [2.0]+[0.0]*(d-1)])
    covP = [jnp.eye(d)]*2
    covQ = covP

    xQ, xP, idx_st, idx_regst = main(prefix="mog2", n=3000, m=300, d=d, pi=pi, mu=mu, covP=covP, covQ=covQ)

    # Unbalanced four modes
    d = 2
    pi = jnp.array([0.1, 0.4, 0.1, 0.4])
    mu = jnp.array([[-3.0, 3.0],
                    [3.0, 3.0],
                    [-3.0, -3.0],
                    [3.0, -3.0]]
    )
    covP = [jnp.eye(d)]*4
    covQ = covP

    xQ, xP, idx_st, idx_regst = main(prefix="mog4", n=3000, m=300, d=d, pi=pi, mu=mu, covP=covP, covQ=covQ)
    print("Stein thinning")
    x_idx = xQ[idx_st]
    w1 = sum((x_idx[:,0]<0) & (x_idx[:,1]>0))/300
    w2 = sum((x_idx[:,0]<0) & (x_idx[:,1]<0))/300
    w3 = sum((x_idx[:,0]>0) & (x_idx[:,1]>0))/300
    w4 = sum((x_idx[:,0]>0) & (x_idx[:,1]<0))/300
    print(f"Weights:", w1, w2, w3, w4)
    print("Regularized Stein thinning")
    x_idx = xQ[idx_regst]
    w1 = sum((x_idx[:,0]<0) & (x_idx[:,1]>0))/300
    w2 = sum((x_idx[:,0]<0) & (x_idx[:,1]<0))/300
    w3 = sum((x_idx[:,0]>0) & (x_idx[:,1]>0))/300
    w4 = sum((x_idx[:,0]>0) & (x_idx[:,1]<0))/300
    print(f"Weights:", w1, w2, w3, w4)

    # Circle
    d = 2
    modes = 6
    radius = 3
    pi = jnp.ones(modes)/modes
    covP = [jnp.eye(d)]*modes
    covQ = covP
    mu = jnp.array([
        [
        np.cos(2.0*np.pi*i/modes)*radius,
        np.sin(2.0*np.pi*i/modes)*radius
        ] for i in range(modes)]
    )

    xQ, xP, idx_st, idx_regst = main(prefix="circle", n=3000, m=300, d=d, pi=pi, mu=mu, covP=covP, covQ=covQ)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
import numpy as np
import kernax
from kernax.utils import median_heuristic
from kernax.toy_mixtures import GaussianMixture
from tqdm import tqdm

def main(m: int, d: int, pi: jnp.ndarray, mu: jnp.ndarray, covP: jnp.ndarray, covQ: jnp.ndarray):

    dist_p = GaussianMixture(d=d, pi=pi, mu=mu, cov=covP)
    dist_q = GaussianMixture(d=d, pi=pi, mu=mu, cov=covQ)

    logprob_fn = dist_p.logprob_fn
    score_fn = jax.grad(logprob_fn)

    np.random.seed(1)
    xQ = dist_q.rand()
    xP = dist_p.rand()
    lengthscale = jnp.array(median_heuristic(xP))

    log_p = jax.vmap(logprob_fn, 0)(xQ)
    score_p = jax.vmap(score_fn, 0)(xQ)
    laplace_log_p = kernax.laplace_log_p_softplus(xQ, logprob_fn)

    st_fn = kernax.SteinThinning(xQ, score_p, lengthscale)
    regst_fn = kernax.RegularizedSteinThinning(xQ, log_p, score_p, laplace_log_p, lengthscale)

    idx_st = st_fn(m)
    idx_regst = regst_fn(m)

    return xQ, xP, idx_st, idx_regst

if __name__ == "__main__":

    # Quantify Pathology I
    d = 2
    pi = jnp.array([0.2, 0.8])
    mu = jnp.array([[-3.0]+[0.0]*(d-1), [3.0]+[0.0]*(d-1)])
    covP = [jnp.eye(d)]*2
    covQ = [jnp.eye(d)]*2

    n = 3000
    m = 300
    w_list = list()
    w_list_rst = list()
    for iter in tqdm(range(100)):
        xQ, xP, idx_st, idx_regst = main(m=m, d=d, pi=pi, mu=mu, covP=covP, covQ=covQ)
        w = sum(xQ[idx_st, 0] < 0)/m
        w_list.append(w)
        w_rst = sum(xQ[idx_regst, 0] < 0)/m
        w_list_rst.append(w_rst)

    # print result for standard Stein thinning
    w_mean = np.mean(w_list)
    w_sd = np.std(w_list)
    print(f"Standard - Left mode weight = {w_mean}, and right mode weight = {1 - w_mean}")
    print(f"Standard - Weight standard deviation = {w_sd}")

    # print result for regularized Stein thinning
    w_mean = np.mean(w_list_rst)
    w_sd = np.std(w_list_rst)
    print(f"Regularized - Left mode weight = {w_mean}, and right mode weight = {1 - w_mean}")
    print(f"Regularized - Weight standard deviation = {w_sd}")


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
import numpy as np

from typing import Callable, Optional
from kernax import kernels
from kernax.utils import median_heuristic
from kernax.types import Array

class SteinThinning:
    """Implements the user interface for the vanilla Stein thinning algorithm that relies on the Langevin Stein operator and the IMQ kernel function.

    Examples
    --------

    Assume that we want to thin a two-dimensional Gaussian output.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (10_000,2))

    We need to define the log-target function.

    .. code::

        from jax.scipy.stats import multivariate_normal

        logprob_fn = multivariate_normal.logpdf

    In this simple example, we could have implemented `logprob_fn` ourselves as follows.

    .. code::

        import jax.numpy as jnp

        logprob_fn = lambda x: -0.5*jnp.sum(x**2)

    Next, we compute a lengthscale and the scores.

    .. code::

        from kernax.utils import median_heuristic
        lengthscale = median_heuristic(x)

        score_fn = lambda x: -x
        score_p = jax.vmap(score_fn, 0)(x)

    We can now get the callable function that performs Stein thinning and compress the Gaussian output.

    .. code::

        stein_thinning_fn = SteinThinning(x, score_p, lengthscale)

        idx = stein_thinning_fn(m = 1_000)

    The selected indices are gathered in the array `idx`.


    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    score_p
        Matrix of size :math:`(n, d)` gathering the score values of :math:`x`
    lengthscale
        Scalar lengthscale that will be used for the underlying IMQ kernel


    Returns
    -------

    A callable function which take a single argument :math:`m` and returns the list of the :math:`m` selected indices.

    """

    def __new__(
        cls,
        x: Array,
        score_p: Array,
        lengthscale: Optional[float]=None
    ) -> Callable:

        if lengthscale is None:
            lengthscale = jnp.array(median_heuristic(x))

        stein_fn = jax.tree_util.Partial(kernels.SteinIMQ, lengthscale=lengthscale)
        # stein_fn = kernels.GetSteinFn(jax.tree_util.Partial(kernels.IMQ, lengthscale=lengthscale))
        kpmap = jax.vmap(stein_fn, (None, None, 0, 0))

        def stein_thinning_fn(m: int) -> Array:
            def step_fn(carry, el):
                obj, idx = carry
                kp = kpmap(x[idx], score_p[idx], x, score_p)
                obj += 2.0*kp
                new_idx = jnp.argmin(obj)
                return (obj, new_idx), new_idx
            init_obj = jax.vmap(lambda x1, s1: stein_fn(x1, s1, x1, s1), (0, 0))(x, score_p)
            init_idx = jnp.argmin(init_obj)
            _, idx = jax.lax.scan(step_fn, (init_obj, init_idx), np.arange(1,m))
            return jnp.append(idx, init_idx)

        return stein_thinning_fn

class RegularizedSteinThinning:
    """Implements the user interface for the regularized Stein thinning algorithm using the Langevin Stein operator and the IMQ kernel function.

    Examples
    --------

    Assume that we want to thin a two-dimensional Gaussian output.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (10_000,2))

    We need to define the log-target function.

    .. code::

        from jax.scipy.stats import multivariate_normal

        logprob_fn = multivariate_normal.logpdf

    In this simple example, we could have implemented `logprob_fn` ourselves as follows.

    .. code::

        import jax.numpy as jnp

        logprob_fn = lambda x: -0.5*jnp.sum(x**2)

    Next, we compute a lengthscale, the scores and laplacian of the log probabilities.

    .. code::

        from kernax.utils import median_heuristic
        lengthscale = median_heuristic(x)

        score_fn = lambda x: -x
        laplace_fn = lambda x: -1.0

        log_p = jax.vmap(logprob_fn, 0)(x)
        score_p = jax.vmap(score_fn, 0)(x)
        laplace_log_p = jax.vmap(laplace_fn, 0)(x)

    Note that the last two quantities can also be computed by automatic differentiation as follows.

    .. code::

        score_fn = jax.grad(logprob_fn)
        score_p = jax.vmap(score_fn, 0)(x)
        laplace_log_p = kernax.laplace_log_p_softplus(x, logprob_fn)

    We can now build the regularized Stein thinning function and call it.

    .. code::

        stein_thinning_fn = SteinThinning(x, log_p, score_p, laplace_log_p, lengthscale)

        idx = stein_thinning_fn(m = 1_000)

    Note that by default, the `RegularizedSteinThinning` uses the median heuristic for the lenthscale if none is passed.

    Parameters
    ----------
    x
        Sample matrix of size :math:`(n, d)`
    log_p
        Callable function built with JAX that returns the log-probability :math:`\log p(x)` of a point :math:`x \in \mathbb{R}^d`.
    score_p
        Optional callable function built with JAX that returns the gradient of :math:`\log p(x)`. Computed with automatic differentiation if not specified by the user.
    laplace_log_p
        Optional callable function built with JAX that returns the laplacian :math:`\Delta \log p(x)`. Computed with automatic differentiation if not specified by the user.
    lengthscale
        Scalar lengthscale that will be used for the underlying IMQ kernel

    Returns
    -------

    A callable function `stein_thinning_fn` which takes two input arguments `m` and `weight_entropy`.

    """

    def __new__(
        cls,
        x: Array,
        log_p: Array,
        score_p: Array,
        laplace_log_p: Array,
        lengthscale: Optional[float]=None
    ) -> Callable:

        if lengthscale is None:
            lengthscale = jnp.array(median_heuristic(x))

        stein_fn = jax.tree_util.Partial(kernels.SteinIMQ, lengthscale=lengthscale)
        # stein_fn = kernels.GetSteinFn(jax.tree_util.Partial(kernels.IMQ, lengthscale=lengthscale))
        kpmap = jax.vmap(stein_fn, (None, None, 0, 0))

        def stein_thinning_fn(m: int, weight_entropy: float=None) -> Array:
            if weight_entropy is None:
                weight_entropy = 1.0/float(m)
            def step_fn(carry, el):
                obj, idx = carry
                kp = kpmap(x[idx], score_p[idx], x, score_p)
                obj += 2.0*kp - weight_entropy*log_p
                new_idx = jnp.argmin(obj)
                return (obj, new_idx), new_idx
            init_obj = jax.vmap(lambda x1, s1: stein_fn(x1, s1, x1, s1), (0, 0))(x, score_p) + laplace_log_p - weight_entropy*log_p
            init_idx = jnp.argmin(init_obj)
            _, idx = jax.lax.scan(step_fn, (init_obj, init_idx), np.arange(1,m))
            return jnp.append(idx, init_idx)

        return stein_thinning_fn


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import os
import jax
import jax.numpy as jnp

import kernax
from kernax.utils import median_heuristic

import numpy as np
from tqdm import tqdm

from absl import app
from absl import logging
from absl import flags

from kernax.toy_mixtures import tBananaMixture

FLAGS = flags.FLAGS
flags.DEFINE_string('algorithm', "mala", 'Name of the MCMC method (mala, nuts)')
flags.DEFINE_string('output_path', '.', 'Path where the results are stored')
flags.DEFINE_integer('dim', 2, 'Dimension of the Gaussian mixture')
flags.DEFINE_integer('num_iterations', 100_000, 'Number of iterations for the mala or nuts algorithm')
flags.DEFINE_integer('thinning_size', 300, 'Number of particles selected by Stein thinning')
flags.DEFINE_integer('num_repetitions', 20, 'Number of repetitions to get uncertainties')
flags.DEFINE_float('step_size', 1e-3, 'Step size for the mala or nuts algorithms')

def main(argv):
    
    d, n, m = FLAGS.dim, FLAGS.num_iterations, FLAGS.thinning_size
    num_reps = FLAGS.num_repetitions
    step_size = FLAGS.step_size
    
    pi = [0.25, 0.75]
    mu = jnp.array([[0.0]*d, [0.0, 8.0]+[0.0]*(d-2)])
    df = 7
    
    dist_p = tBananaMixture(d=d, pi=pi, mu=mu, df=df)
    
    logprob_fn = dist_p.logprob_fn
    score_fn = jax.grad(logprob_fn)
    
    if FLAGS.algorithm=="nuts":
        inverse_mass_matrix = jnp.ones(d)
        def sample_fn(init_key):
            init_positions = jax.random.normal(init_key, (d,))
            _, rng_key = jax.random.split(init_key)
            _, states = kernax.nuts(logprob_fn, init_positions, n, step_size, inverse_mass_matrix, rng_key)
            return states.position
    elif FLAGS.algorithm=="mala":
        def sample_fn(init_key):
            init_positions = jax.random.normal(init_key, (d,))
            _, rng_key = jax.random.split(init_key)            
            _, states = kernax.mala(logprob_fn, init_positions, n, step_size, rng_key)
            return states.position
    
    rng_key = jax.random.PRNGKey(np.random.randint(0,1234567))
    rep_keys = jax.random.split(rng_key, num_reps)
        
    idx, ridx = [], []
    samples = []
    for key in tqdm(rep_keys, leave=True):
        
        logging.info(f"Sampling with {FLAGS.algorithm}")
        positions = sample_fn(key)    
        
        logging.info(f"Computing log-prob, scores, and laplacians")
        log_p = jax.vmap(logprob_fn, 0)(positions)
        score_p = jax.vmap(score_fn, 0)(positions)
        laplace_log_p = kernax.laplace_log_p_softplus(positions, logprob_fn)
        lengthscale = median_heuristic(positions)
    
        logging.info("Stein thinning")
        stein_thinning_fn = kernax.SteinThinning(positions, score_p, lengthscale)
        idx += [stein_thinning_fn(m=m)]
                
        logging.info("Regularized Stein thinning")
        stein_thinning_fn = kernax.RegularizedSteinThinning(positions, log_p, score_p, laplace_log_p, lengthscale)
        ridx += [stein_thinning_fn(m=m)]
        
        samples += [positions]
            
    np.savez(os.path.join(FLAGS.output_path, f"mob_{FLAGS.algorithm}_sz_{FLAGS.step_size}_m_{FLAGS.thinning_size}_d_{FLAGS.dim}.npz"), idx=idx, ridx=ridx, samples=samples, pi=pi, mu=mu, df=df)
            
if __name__ == "__main__":
    
    app.run(main)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import os
import jax.numpy as jnp
import numpy as np

import matplotlib.pyplot as plt
try:
    plt.rcParams.update({
            "text.usetex": True,
            "font.family": "sans-serif",
            "font.sans-serif": ["Helvetica"]})
    plt.style.use('seaborn-v0_8')
    FONTSIZE=24
except:
    FONTSIZE=20

def rand_gaussian_mixture(n: int, pi: list, mu, cov):
    x = []
    for _ in range(n):
        z_i = np.argmax(np.random.multinomial(1, pi))
        x_i = np.random.multivariate_normal(mu[z_i], cov[z_i], size=1).T
        x.append(x_i)
    return np.array(x).squeeze()

if __name__ == "__main__":

    dimensions = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    nd = len(dimensions)

    for ai, algorithm in enumerate(["mala", "nuts"]):

        subfig, axs = plt.subplots(1, 3, figsize=(3*6,4), sharex=True, sharey="row")
        markers = ["o", "s"]
        labels = [r"$\mathrm{Stein}$ $\mathrm{Thinning}$", r"$\mathrm{Regularized}$ $\mathrm{Stein}$ $\mathrm{Thinning}$"]

        for si, step_size in enumerate(["0.05", "0.1", "0.5"]):

            mean_mmd_st, std_mmd_st = np.zeros(nd), np.zeros(nd)
            mean_mmd_rst, std_mmd_rst = np.zeros(nd), np.zeros(nd)

            for i, dimension in enumerate(dimensions):
                data = jnp.load(os.path.join("mmd_data", f"metrics_{algorithm}_sz_{step_size}_m_300_d_{dimension}.npz"))

                mean_mmd_st[i] = data["mean_mmd_st"]
                std_mmd_st[i] = data["std_mmd_st"]

                mean_mmd_rst[i] = data["mean_mmd_rst"]
                std_mmd_rst[i] = data["std_mmd_rst"]

            fig, ax = plt.subplots(1, 1, figsize=(6,4))
            ax.plot(dimensions, mean_mmd_st, linewidth=2, linestyle="-", color="b", marker="o", label=labels[0])
            ax.plot(dimensions, mean_mmd_rst, linewidth=2, linestyle="-", color="r", marker="o", label=labels[1])
            ax.fill_between(dimensions, mean_mmd_st-std_mmd_st, mean_mmd_st+std_mmd_st, color="tab:blue", alpha=0.25)
            ax.fill_between(dimensions, mean_mmd_rst-std_mmd_rst, mean_mmd_rst+std_mmd_rst, color="tab:red", alpha=0.25)
            ax.set_xlabel(r"$d$", fontsize=24)
            ax.set_ylabel(r"$\mathrm{MMD}(P,Q_m)$", fontsize=24)
            ax.tick_params(labelsize=22)
            ax.set_xticks(dimensions)
            fig.savefig(f"mog_{algorithm}_sz_{step_size}_dimension.png", format="png", bbox_inches="tight")

            axs[si].plot(dimensions, mean_mmd_st, linewidth=2, linestyle="-", color="b", marker="o", label=labels[0])
            axs[si].plot(dimensions, mean_mmd_rst, linewidth=2, linestyle="-", color="r", marker="o", label=labels[1])
            axs[si].fill_between(dimensions, mean_mmd_st-std_mmd_st, mean_mmd_st+std_mmd_st, color="tab:blue", alpha=0.25)
            axs[si].fill_between(dimensions, mean_mmd_rst-std_mmd_rst, mean_mmd_rst+std_mmd_rst, color="tab:red", alpha=0.25)
            axs[si].set_xlabel(r"$d$", fontsize=20)
            axs[si].set_ylabel(r"$\mathrm{MMD}(P,Q_m)$", fontsize=20)
            axs[si].tick_params(labelsize=18)
            axs[si].set_xticks(dimensions)
            axs[si].set_title(rf"$\epsilon = {step_size}$", fontsize=18)

        handles, labels = axs[0].get_legend_handles_labels()
        subfig.legend(handles, labels, loc=(0.12, 0.50), ncols=1, fontsize=20)
        subfig.savefig(f"subplot_mog_{algorithm}_dimension.png", format="png", bbox_inches="tight")


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import numpy as np
import jax.numpy as jnp
from jax.scipy.stats import multivariate_normal
from scipy.stats import multivariate_t

from typing import NamedTuple, Callable, List
from functools import partial

def _rand_gaussian_mixture(n: int, pi: list, mu, cov):
    x = []
    for _ in range(n):
        z_i = np.argmax(np.random.multinomial(1, pi))
        x_i = np.random.multivariate_normal(mu[z_i], cov[z_i], size=1).T
        x.append(x_i)
    return jnp.array(x).squeeze()

def _rand_gaussian_banana(mu: np.ndarray, cov: np.ndarray):
    d = len(mu)
    b = 0.025
    z = np.random.multivariate_normal(np.zeros(d), np.diag(cov), size=1).T
    x = np.zeros_like(z)
    x = z
    x[1] += b*z[0]**2 - 100*b
    y = x.squeeze() + mu
    return y

def _rand_gaussian_banana_mixture(mu: np.ndarray, pi: np.ndarray, cov_diag: np.ndarray, size: int=10_000):
    _, d = mu.shape
    x = np.zeros((size,d))
    for i in range(size):
        zi = np.argmax(np.random.multinomial(1, pi))
        x[i] = _rand_gaussian_banana(mu[zi], cov_diag)
    return x

def _rand_t_banana(mu: np.ndarray, cov: np.ndarray, df: int):
    d = len(mu)
    b = 0.025
    z = multivariate_t.rvs(np.zeros(d), cov, df=df, size=1)
    x = np.zeros_like(z)
    x = z
    x[1] += b*z[0]**2 - 100*b
    return x + mu

def _rand_t_banana_mixture(mu: np.ndarray, pi: np.ndarray, cov_diag: np.ndarray, df: int, size: int=10_000):
    _, d = mu.shape
    x = np.zeros((size,d))
    for i in range(size):
        zi = np.argmax(np.random.multinomial(1, pi))
        x[i] = _rand_t_banana(mu[zi], np.diag(cov_diag), df)
    return x

def _logpdf_multivariate_t(x, loc, df, prec_U, dim):
    dev = x - loc
    maha = jnp.sum(jnp.square(dev*prec_U))
    t = 0.5 * (df + dim)
    logpdf = -t * jnp.log(1. + (1./df) * maha) + 2*dim
    return logpdf

def _pdf_multivariate_t(x, loc, df, prec_U, dim):
    log_pdf = _logpdf_multivariate_t(x, loc, df, prec_U, dim)
    return jnp.exp(log_pdf)

class Distribution(NamedTuple):
    d: int
    logprob_fn: Callable
    rand: Callable

class GaussianMixture:

    def __new__(
        cls,
        d: int,
        pi: List[int],
        mu: jnp.ndarray,
        cov: jnp.ndarray) -> Distribution:

        def logprob_fn(x: jnp.ndarray):
            K = 1.5*d
            pdf = pi[0]*jnp.exp(K + multivariate_normal.logpdf(x, mu[0], cov[0]))
            for i in range(1,len(pi)):
                pdf += pi[i]*jnp.exp(K + multivariate_normal.logpdf(x, mu[i], cov[i]))
            return jnp.log(pdf)

        rand = partial(_rand_gaussian_mixture, pi=pi, mu=mu, cov=cov)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)

class gBananaMixture:
    """Gaussian banana mixture.
    """

    def __new__(
        cls,
        d: int,
        pi: List[int],
        mu: List[np.ndarray]
        ) -> Distribution:

        dim = d
        cov_diag = np.array([30] + [1]*(d-1))

        def logprob_fn(x: np.ndarray):
            b = 0.025
            z = x-mu[0]
            z = z.at[1].set(x[1]-b*z[0]**2 + 100*b - mu[0][1])
            pdf = pi[0]*multivariate_normal.pdf(z, jnp.zeros(dim), jnp.diag(cov_diag))
            for i in range(1,len(pi)):
                z = x-mu[i]
                z = z.at[1].set(x[1]-b*z[0]**2 + 100*b - mu[i][1])
                pdf += pi[i]*multivariate_normal.pdf(z, jnp.zeros(dim), jnp.diag(cov_diag))
            return jnp.log(pdf)

        rand = partial(_rand_gaussian_banana_mixture, mu=mu, pi=pi, cov_diag=cov_diag)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)

class tBananaMixture:
    """Student-t banana mixture.
    """

    def __new__(
        cls,
        d: int,
        pi: List[int],
        mu: List[np.ndarray],
        df: int) -> Distribution:

        dim = d
        cov_diag = np.array([30] + [1]*(d-1))
        prec_U = jnp.sqrt(1.0/cov_diag)

        def logprob_fn(x: np.ndarray):
            b = 0.025
            z = x-mu[0]
            z = z.at[1].set(x[1]-b*z[0]**2 + 100*b - mu[0][1])
            pdf = pi[0]*_pdf_multivariate_t(z, jnp.zeros(dim), df, prec_U, dim)
            for i in range(1,len(pi)):
                z = x-mu[i]
                z = z.at[1].set(x[1]-b*z[0]**2 + 100*b - mu[i][1])
                pdf += pi[i]*_pdf_multivariate_t(z, jnp.zeros(dim), df, prec_U, dim)
            return jnp.log(pdf)

        rand = partial(_rand_t_banana_mixture, mu=mu, pi=pi, cov_diag=cov_diag, df=df)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# sys.path.insert(0, os.path.abspath('../batorch'))
# sys.path.insert(0, os.path.abspath('..'))
basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, basedir)
sys.path.insert(0, os.path.join(basedir, "kernax"))
print(sys.path)

import kernax
project = 'Kernax'
copyright = 'Safran Group'
author = 'Safran Group'

release = '0.0.1'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    "sphinx.ext.intersphinx",
    "sphinx_math_dollar",
    "sphinx.ext.mathjax",
    "myst_nb",    
]

numfig = True
autosummary_generate = True

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
html_theme = 'sphinx_book_theme'
html_css_files = ['custom.css']
html_static_path = ["_static"]
html_title = "Kernax"
html_css_files = ["custom.css"]
html_context = {
   "default_mode": "light"
}

autodoc_mock_imports = ["jax"]  #"jax.numpy", "kernax", "kernax.thinning", "kernax.discrepancies", "kernax.kernels"]
autodoc_member_order = 'bysource'


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax.numpy as jnp
from blackjax.types import PyTree, PRNGKey

#: JAX Array
Array = jnp.ndarray

#: JAX PyTrees
PyTree = PyTree

#: JAX PRNGKey
PRNGKey = PRNGKey


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import os
import numpy as np
import pandas as pd

def GermanCredit(path_to_folder: str):
    """Loads the UCI German Credit dataset.
    
    https://archive.ics.uci.edu/ml/datasets/statlog+(german+credit+data)
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file haberman.data is located
        
    Returns
    -------
    
    Tuple (X, y)
    
    """
    
    data = np.genfromtxt(os.path.join(path_to_folder, "german.data-numeric"))
    X = data[:,:-1]
    y = data[:,-1:].squeeze()
    return X, y

def HabermanSurvival(path_to_folder: str):
    """Loads the UCI Haberman dataset.
    
    https://archive.ics.uci.edu/ml/datasets/haberman%27s+survival
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file haberman.data is located
        
    Returns
    -------
    
    Tuple (X, y)
    
    """
    
    data = np.genfromtxt(os.path.join(path_to_folder, "haberman.data"), delimiter=",")
    X = data[:,:-1]
    y = data[:,-1:].squeeze() - 1
    return X, y

def LiverDisorder(path_to_folder: str):
    """Loads the UCI Liver disorder dataset.
    
    https://archive.ics.uci.edu/ml/datasets/liver+disorders
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file bupa.data is located
        
    Returns
    -------
    
    Tuple (X, y)
    
    """
    
    data = np.genfromtxt(os.path.join(path_to_folder, "bupa.data"), delimiter=",")
    X = data[:,:-1]
    y = data[:,-1:].squeeze() - 1
    return X, y

def BreastCancer():
    """Loads the breast cancer dataset with sklearn directly.
    
    Returns
    -------
    
    Tuple (X, y)
    
    """
    
    from sklearn.datasets import load_breast_cancer
    X, y = load_breast_cancer(return_X_y=True)
    return X, y

def Authentification(path_to_folder: str):
    """Loads the UCI authentification dataset.
    
    https://archive.ics.uci.edu/ml/datasets/banknote+authentication
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file data_banknote_authentication.txt is located
        
    Returns
    -------
    
    Tuple (X, y)    
    
    """
    
    data = np.genfromtxt(os.path.join(path_to_folder, "data_banknote_authentication.txt"), delimiter=",")
    X = data[:,:-1]
    y = data[:,-1:].squeeze()
    return X, y

def Ionosphere(path_to_folder: str):
    """Loads the UCI ionosphere dataset.
    
    https://archive.ics.uci.edu/ml/datasets/ionosphere
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file ionosphere.data is located
        
    Returns
    -------
    
    Tuple (X, y)    
    
    """
    
    # data = np.genfromtxt(os.path.join(path_to_folder, "ionosphere.data"), delimiter=",")
    data = pd.read_csv(os.path.join(path_to_folder, "ionosphere.data"))
    X = data.iloc[:,:-1].values
    y_raw = data.iloc[:,-1:].values
    def map_fn(y):
        if y=='b':
            return 1
        elif y=="g":
            return 0
    y = [map_fn(yi) for yi in y_raw]
    return X, np.array(y)

def Sonar(path_to_folder: str):
    """Loads the UCI Sonar dataset.
    
    http://archive.ics.uci.edu/ml/datasets/connectionist+bench+(sonar,+mines+vs.+rocks)
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file sonar.all-data is located
        
    Returns
    -------
    
    Tuple (X, y)    
    
    """
    
    data = pd.read_csv(os.path.join(path_to_folder, "sonar.all-data"))
    X = data.iloc[:,:-1].values
    y_raw = data.iloc[:,-1:].values
    def map_fn(y):
        if y=='R':
            return 1
        elif y=="M":
            return 0
    y = [map_fn(yi) for yi in y_raw]
    return X, np.array(y)

def Diabetes(path_to_folder: str):
    """Loads the UCI Diabetes dataset.
    
    Parameters
    ----------
    
    path_to_folder
        Path to the folder where the file diabetes.data is located
        
    Returns
    -------
    
    Tuple (X, y)
    
    """
    
    data = pd.read_csv(os.path.join(path_to_folder, "diabetes.data"))
    X = data.iloc[:,:-1].values
    y = data.iloc[:,-1:].values
    return X, y


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import multiprocessing
import os

os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count={}".format(
    multiprocessing.cpu_count()
)

import os
import jax
import jax.numpy as jnp
import numpy as np
import kernax
from kernax.utils import median_heuristic, laplace_log_p_softplus

from absl import app
from absl import flags
from absl import logging
from uci_datasets import BreastCancer, Ionosphere, Sonar, Authentification, HabermanSurvival, LiverDisorder, Diabetes
from sklearn.model_selection import KFold
from sklearn.metrics import roc_curve, auc
from tqdm import tqdm

FLAGS = flags.FLAGS
flags.DEFINE_string('dataset_name', None, 'Name of the dataset (Ionosphere, Sonar, Authentification, BreastCancer)')
flags.DEFINE_string('path_to_data', None, 'Path to the folder where the data is stored')
flags.DEFINE_string('algorithm', None, 'Name of the MCMC method (mala, nuts)')
flags.DEFINE_float('step_size', None, 'Step size for the mala or nuts algorithms')
flags.DEFINE_integer('num_iterations', 1000, 'Number of iterations for the mala or nuts algorithm')
flags.DEFINE_integer('thinning_size', 300, 'Number of particles selected by Stein thinning')
flags.DEFINE_integer('num_repetitions', 10, 'Number of repitions to get uncertainties')
flags.DEFINE_string('output_path', ".", 'Folder where the outputs are dumped to')

def logprior_fn(params):
    """t-Student log prior function"""
    a, b = 1.0, 1.0
    c = (2.0*a + 1.0)/2.0
    logprobs = c*jnp.log(1.0 + jnp.square(params)/(2.0*b))
    return -jnp.sum(logprobs)

def loglikelihood_fn(params, data):
    """Log likelihood function for a given observation"""
    x, y = data
    logit = jnp.dot(x, params)
    sigmoid = jax.nn.sigmoid(logit)
    # return y*jnp.log(sigmoid) + (1-y)*jnp.log(1-sigmoid)
    return y*sigmoid - jnp.log1p(jnp.exp(sigmoid))

def predict_fn(w, x):
    """Predict function for logistic regression for a given weight w and a given input x"""
    logits = jnp.dot(w, x)
    return jax.nn.sigmoid(logits)
    
def init_parameters(rng_key, X_train):
    w = jax.random.normal(rng_key, (X_train.shape[1],))
    return w    
    
def logprob_fn(parameters, data):
    X_train, y_train = data
    logprior = logprior_fn(parameters)
    batch_loglikelihood = jax.vmap(loglikelihood_fn, (None, 0))(parameters, (X_train, y_train))
    return logprior + jnp.sum(batch_loglikelihood)

def sample_mala(rng_key, step_size, num_samples, X_train, y_train):
    # Get data and log-posterior
    log_posterior_fn = jax.tree_util.Partial(logprob_fn, data=(X_train, y_train))    

    # Initialize parameters
    init_positions = init_parameters(rng_key, X_train)

    # Run MALA
    _, rng_key = jax.random.split(rng_key)
    _, states = kernax.mala(log_posterior_fn, init_positions, num_samples, step_size, rng_key)
    return states.position
    
def sample_nuts(rng_key, step_size, num_samples, X_train, y_train):
    # Get data and log-posterior
    log_posterior_fn = jax.tree_util.Partial(logprob_fn, data=(X_train, y_train))
    
    # Initialize parameters
    inverse_mass_matrix = jnp.ones(X_train.shape[1])
    init_positions = init_parameters(rng_key, X_train)

    # Run NUTS
    _, rng_key = jax.random.split(rng_key)
    _, states = kernax.nuts(log_posterior_fn, init_positions, num_samples, step_size, inverse_mass_matrix, rng_key)
    return states.position

def main(argv):
    
    logging.info(f"Number of devices: {len(jax.devices())}")
    num_chains = len(jax.devices())
    
    step_size = FLAGS.step_size
    num_iterations = FLAGS.num_iterations
    m = FLAGS.thinning_size
    
    if FLAGS.dataset_name=="Ionosphere":
        X, y = Ionosphere(FLAGS.path_to_data)
    elif FLAGS.dataset_name=="Authentification":
        X, y = Authentification(FLAGS.path_to_data)
    elif FLAGS.dataset_name=="Sonar":
        X, y = Sonar(FLAGS.path_to_data)
    elif FLAGS.dataset_name=="BreastCancer":
        X, y = BreastCancer()
    elif FLAGS.dataset_name=="HabermanSurvival":
        X, y = HabermanSurvival(FLAGS.path_to_data)
    elif FLAGS.dataset_name=="LiverDisorder":
        X, y = LiverDisorder(FLAGS.path_to_data)
    elif FLAGS.dataset_name=="Diabetes":
        X, y = Diabetes(FLAGS.path_to_data)
    else:
        raise ValueError(f"Wrong dataset name, I got {FLAGS.dataset_name}")    
    
    logging.info(f"Dataset: {FLAGS.dataset_name}")
    logging.info(f"y.shape: {y.shape}")
    
    auc_st = []
    auc_rst = []
    for _ in tqdm(range(FLAGS.num_repetitions)):
        
        kf = KFold(n_splits=10, shuffle=True)
        
        mean_prob_st = []
        mean_prob_rst = []
        y_test = []
        
        for train_index, test_index in kf.split(X, y):
            
            X_train = jnp.array(X[train_index])
            y_train = jnp.array(y[train_index])
            X_test = jnp.array(X[test_index])
            y_test += [y[test_index]]
                
            score_fn = jax.grad(logprob_fn, argnums=0)
            
            logging.info(f"Sampling with {FLAGS.algorithm}")
            if FLAGS.algorithm=="mala":
                filename = f"logistic_regression_{FLAGS.dataset_name}_mala_sz_{step_size}_m_{FLAGS.thinning_size}.npz"
                
                pmap_mcmc = jax.pmap(sample_mala, in_axes=(0, None, None, None, None), out_axes=(0), static_broadcasted_argnums=(2,))
                rng_key = jax.random.PRNGKey(np.random.randint(0,124567))
                rng_keys = jax.random.split(rng_key, num_chains)
                position = pmap_mcmc(rng_keys, step_size, num_iterations, X_train, y_train)
                _ = jax.tree_util.tree_map(lambda x: x.block_until_ready(), position)
                position = jnp.reshape(position, (num_chains*num_iterations,-1))
                        
            elif FLAGS.algorithm=="nuts":
                filename = f"logistic_regression_{FLAGS.dataset_name}_nuts_sz_{step_size}_m_{FLAGS.thinning_size}.npz"
                
                pmap_mcmc = jax.pmap(sample_nuts, in_axes=(0, None, None, None, None), out_axes=(0), static_broadcasted_argnums=(2,))
                rng_key = jax.random.PRNGKey(np.random.randint(0,124567))
                rng_keys = jax.random.split(rng_key, num_chains)
                position = pmap_mcmc(rng_keys, step_size, num_iterations, X_train, y_train)
                _ = jax.tree_util.tree_map(lambda x: x.block_until_ready(), position)
                position = jnp.reshape(position, (num_chains*num_iterations,-1))
                
            else:
                raise ValueError("Wrong algorithm")
            
            logging.info("Computing log-probs, scores, and laplacians")
            log_p = jax.vmap(logprob_fn, (0, None))(position, (X_train, y_train))
            score_p = jax.vmap(score_fn, (0, None))(position, (X_train, y_train))
            
            log_posterior_fn = jax.tree_util.Partial(logprob_fn, data=(X_train, y_train))
            laplace_log_p = laplace_log_p_softplus(position, log_posterior_fn)
            
            subidx = np.random.randint(0,len(position),size=48_000)
            lengthscale = median_heuristic(position[subidx])
            
            logging.info("Stein thinning")
            stein_thinning_fn = kernax.SteinThinning(position, score_p, lengthscale)
            idx = stein_thinning_fn(m)
            
            logging.info("Predicting")
            params_st = position[idx]
            prob_st = jnp.stack([jax.vmap(predict_fn, in_axes=(None, 0))(params_st[i], X_test) for i in range(len(params_st))])
            mean_prob_st += [jnp.mean(prob_st, 0)]
            
            logging.info("Regularized Stein thinning")
            reg_stein_thinning_fn = kernax.RegularizedSteinThinning(position, log_p, score_p, laplace_log_p, lengthscale)
            reg_idx = reg_stein_thinning_fn(m)
            
            logging.info("Predicting")
            params_rst = position[reg_idx]
            prob_rst = jnp.stack([jax.vmap(predict_fn, in_axes=(None, 0))(params_rst[i], X_test) for i in range(len(params_rst))])
            mean_prob_rst += [jnp.mean(prob_rst, 0)]

        mean_prob_st = jnp.concatenate(mean_prob_st)
        mean_prob_rst = jnp.concatenate(mean_prob_rst)
        y_test = np.concatenate(y_test)
        
        logging.info("Computing ROC and AUC")
        fpr_st, tpr_st, _ = roc_curve(y_test, mean_prob_st, drop_intermediate=False)
        auc_st += [auc(fpr_st, tpr_st)]
        
        logging.info("Computing ROC and AUC")
        fpr_rst, tpr_rst, _ = roc_curve(y_test, mean_prob_rst, drop_intermediate=False)
        auc_rst += [auc(fpr_rst, tpr_rst)]
        
    mean_auc_st = np.mean(auc_st)
    std_auc_st = np.std(auc_st)
    
    mean_auc_rst = np.mean(auc_rst)
    std_auc_rst = np.std(auc_rst)
    
    logging.info(f"Saving to disk: {os.path.join(FLAGS.output_path, filename)}")
    jnp.savez(os.path.join(FLAGS.output_path, filename), mean_auc_st=mean_auc_st, std_auc_st=std_auc_st, mean_auc_rst=mean_auc_rst, std_auc_rst=std_auc_rst) 

if __name__ == "__main__":
    
    app.run(main)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

try:
    from ._version import __version__
except ImportError:
    __version__ = "None"

from .thinning import (
    SteinThinning,
    RegularizedSteinThinning
)

from .discrepancies import (
    MMD,
    KSD
)

from .bjsamplers import (
    rmh,
    hmc,
    nuts,
    mala
)

from .utils import (
    laplace_log_p_hardplus,
    laplace_log_p_softplus
)

__all__ = ["__version__",
           "SteinThinning",
           "RegularizedSteinThinning",
           "laplace_log_p_hardplus",
           "laplace_log_p_softplus",
           "MMD",
           "KSD",
           "rmh",
           "hmc",
           "nuts",
           "mala"]


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

# import os
# import re
from setuptools import setup

# VERSIONFILE="kernax/_version.py"
# # if os.path.isfile(VERSIONFILE):
# verstrline = open(VERSIONFILE, "rt").read()
# VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
# mo = re.search(VSRE, verstrline, re.M)
# if mo:
#     verstr = mo.group(1)
# else:
#     raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
    name="kernax",
    # version=verstr,
    packages=["kernax"],
    description="Kernelized Stein discrepancy thinning",
    author="Brian Staber",
    author_email="brian.staber@safrangroup.com",
    include_package_data=True
)



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
import numpy as np
from scipy.spatial.distance import pdist

from typing import Callable
from kernax.types import Array

def median_heuristic(x: np.ndarray):
    """Function that computes the median heuristic for the lengthscale parameter.

    Parameters
    ----------

    x
        Sample matrix of dimensions :math:`(n \times d)`.

    Returns
    -------

    Median of the pairwise distances.
    """
    return np.median(pdist(x))

def laplace_log_p_hardplus(x: Array, logprob_fn: Callable) -> Array:
    """Function that computes the clipped laplacian of a log-probability for the provided sample matrix.

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    logprob_fn
        Callable log-probability function

    Returns
    -------

    Values of the laplacian of log-probability.
    """
    jacobian_fn = jax.jacfwd(jax.jacrev(logprob_fn))
    jacobians = jax.vmap(jacobian_fn, 0)(x)
    llp = jnp.nan_to_num(jnp.clip(jax.vmap(jnp.trace)(jacobians),a_min=0))
    return llp

def laplace_log_p_softplus(x: Array, logprob_fn: Callable) -> Array:
    """Function that computes the sum of positive second-order derivatives of the log-probability for the provided sample matrix.

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    logprob_fn
        Callable log-probability function

    Returns
    -------

    Values of the laplacian of log-probability.
    """
    jacobian_fn = jax.jacfwd(jax.jacrev(logprob_fn))
    jacobians = jnp.nan_to_num(jnp.clip(jax.vmap(jacobian_fn, 0)(x), a_min=0))
    llp = jax.vmap(jnp.trace)(jacobians)
    return llp

def laplace_log_p_flax(n: int, x: Array, logprob_fn: Callable) -> Array:
    """Function that computes the sum of positive second-order derivatives of the log-probability for the provided sample matrix.

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    logprob_fn
        Callable log-probability function

    Returns
    -------

    Values of the laplacian of log-probability.
    """
    layers = x.keys()
    jacobian_fn = jax.jacfwd(jax.jacrev(logprob_fn))
    jacobians = jax.vmap(jacobian_fn, 0)(x)
    laplacian = jnp.zeros(n)
    for layer in layers:
        jkernel = jnp.nan_to_num(jnp.clip(jacobians[layer]["kernel"][layer]["kernel"], a_min=0))
        jbias = jnp.nan_to_num(jnp.clip(jacobians[layer]["bias"][layer]["bias"], a_min=0))
        laplacian += jax.vmap(lambda xx: jnp.trace(jnp.reshape(xx, (xx.shape[0]*xx.shape[1],xx.shape[2]*xx.shape[3]))))(jkernel) + jax.vmap(jnp.trace,0)(jbias)
    return laplacian


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp

import blackjax
from blackjax.mcmc.hmc import HMCState
from blackjax.mcmc.mala import MALAState
from kernax.types import Array, PRNGKey

from typing import Callable

def inference_loop(rng_key, step_fn, initial_state, num_samples):
    """Inference loop with lax.scan"""
    @jax.jit
    def one_step(state, rng_key):
        state, info = step_fn(rng_key, state)
        return state, (info, state)
    keys = jax.random.split(rng_key, num_samples)
    _, (infos, states) = jax.lax.scan(one_step, initial_state, keys)
    return infos, states

def rmh(logprob_fn: Callable,
        init_positions: Array,
        num_samples: int,
        sigma: Array,
        rng_key: PRNGKey) -> HMCState:
    """Wrapper of the HMC algorithm implemented in BlackJAX.

    Parameters
    ----------

    logprob_fn
        Callable function that returns the log-probability of the target
    init_positions
        Initial guess for the HMC algorithm
    num_samples
        Number of iterations (burn-in included)
    sigma
        Scaling of the Gaussian proposal (vector)
    rng_key
        A JAX PRNGKey.

    Returns
    -------

    A tuple of states and informations
    """

    kern = blackjax.rmh(logprob_fn, sigma)
    step_fn = jax.jit(kern.step)
    init_state = kern.init(init_positions)
    infos, states = inference_loop(rng_key, step_fn, init_state, num_samples)
    return infos, states

def hmc(logprob_fn: Callable,
        init_positions: Array,
        num_samples: int,
        step_size: float,
        inverse_mass_matrix: Array,
        num_integration_steps: int,
        rng_key: PRNGKey) -> HMCState:
    """Wrapper of the HMC algorithm implemented in BlackJAX.

    Parameters
    ----------

    logprob_fn
        Callable function that returns the log-probability of the target
    init_positions
        Initial guess for the HMC algorithm
    num_samples
        Number of iterations (burn-in included)
    inverse_mass_matrix
        Flattened inverse mass matrix
    num_integration_steps
        Number of integration steps
    rng_key
        A JAX PRNGKey.

    Returns
    -------

    A tuple of states and informations
    """

    kern = blackjax.hmc(logprob_fn, step_size, inverse_mass_matrix, num_integration_steps)
    step_fn = jax.jit(kern.step)
    init_state = kern.init(init_positions)
    infos, states = inference_loop(rng_key, step_fn, init_state, num_samples)
    return infos, states

def nuts(logprob_fn: Callable,
         init_positions: Array,
         num_samples: int,
         step_size: float,
         inverse_mass_matrix: Array,
         rng_key: PRNGKey) -> HMCState:
    """Wrapper of the NUTS algorithm implemented in BlackJAX.

    Parameters
    ----------

    logprob_fn
        Callable function that returns the log-probability of the target
    init_positions
        Initial guess for the HMC algorithm
    num_samples
        Number of iterations (burn-in included)
    step_size
        Step size of the leapfrog integrator
    inverse_mass_matrix
        Flattened inverse mass matrix
    rng_key
        A JAX PRNGKey.

    Returns
    -------

    A tuple of states and informations
    """

    kern = blackjax.nuts(logprob_fn, step_size, inverse_mass_matrix)
    step_fn = jax.jit(kern.step)
    init_state = kern.init(init_positions)
    infos, states = inference_loop(rng_key, step_fn, init_state, num_samples)
    return infos, states

def mala(logprob_fn: Callable,
         init_positions: Array,
         num_samples: int,
         step_size: float,
         rng_key: PRNGKey) -> MALAState:
    """Wrapper of the MALA algorithm implemented in BlackJAX.

    Parameters
    ----------

    logprob_fn
        Callable function that returns the log-probability of the target
    init_positions
        Initial guess for MALA
    num_samples
        Number of iterations (burn-in included)
    step_size
        Step size of the integrator
    rng_key
        A JAX PRNGKey.

    Returns
    -------

    A tuple of states and informations
    """

    kern = blackjax.mala(logprob_fn, step_size)
    step_fn = jax.jit(kern.step)
    init_state = kern.init(init_positions)
    infos, states = inference_loop(rng_key, step_fn, init_state, num_samples)
    return infos, states


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import os
import jax.numpy as jnp

if __name__ == "__main__":

    algorithms = ["mala"]
    datasets = ["BreastCancer", "Diabetes", "HabermanSurvival", "LiverDisorder", "Sonar"]
    step_sizes = [0.01, 0.001, 0.0001, 1e-05]
    thinning_sizes = [50, 100, 300]

    for algorithm in algorithms:
        auc_mean_st = {}

        for dataset in datasets:
            auc_mean_st[dataset] = {}

            for thinning_size in thinning_sizes:
                mean_auc_st, mean_auc_rst = [], []
                std_auc_st, std_auc_rst = [], []
                for step_size in step_sizes:
                    filename = os.path.join("mcmc_data", f"logistic_regression_{dataset}_{algorithm}_sz_{step_size}_m_{thinning_size}.npz")
                    if os.path.isfile(filename):
                        # data = jnp.load(os.path.join("22_11_2022", filename))
                        data = jnp.load(filename)

                        mean_auc_st += [data["mean_auc_st"]]
                        mean_auc_rst += [data["mean_auc_rst"]]
                        std_auc_st += [data["std_auc_st"]]
                        std_auc_rst += [data["std_auc_rst"]]
                    else:
                        raise ValueError(f"Missing {filename}")

                ist = jnp.argmax(jnp.array(mean_auc_st))
                irst = jnp.argmax(jnp.array(mean_auc_rst))
                auc_mean_st[dataset][str(thinning_size)] = "{:1.2f} ({:1.3f}) {:1.2f} ({:1.3f})".format(round(mean_auc_st[ist].item(),2), round(std_auc_st[ist].item(),4), round(mean_auc_rst[ist].item(),2), round(std_auc_rst[ist].item(),4))

        import pandas as pd
        df = pd.DataFrame.from_dict(auc_mean_st)
        print(df.T)
        print("*"*100)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax
import jax.numpy as jnp
import numpy as np

import os
from kernax import MMD
from tqdm import tqdm

from absl import app
from absl import flags

FLAGS = flags.FLAGS
flags.DEFINE_string('algorithm', "mala", 'Name of the MCMC method (mala, nuts)')
flags.DEFINE_float('step_size', None, 'Step size for the mala or nuts algorithms')
flags.DEFINE_integer('thinning_size', None, 'Thinning size')
flags.DEFINE_integer('dimension', 2, 'Dimension')

from kernax.toy_mixtures import tBananaMixture

def main(argv):

    data = jnp.load(os.path.join("mcmc_data", f"mob_{FLAGS.algorithm}_sz_{FLAGS.step_size}_m_{FLAGS.thinning_size}_d_{FLAGS.dimension}.npz"))
    pi, mu, df = data["pi"], data["mu"], data["df"]
    samples, idx, ridx = data["samples"], data["idx"], data["ridx"]

    num_reps = len(samples)
    mmd_wrt_m_st = np.zeros(num_reps)
    mmd_wrt_m_rst = np.zeros(num_reps)

    dist_p = tBananaMixture(d=FLAGS.dimension, pi=pi, mu=mu, df=df)

    for i in tqdm(range(num_reps), leave=True, desc="reps"):

        xq, idx_st, idx_rst = samples[i], idx[i], ridx[i]
        xp = dist_p.rand()

        xx = xq[idx_st]
        yy = xq[idx_rst]

        mmd_wrt_m_st[i] = MMD(x=xp, y=xx)
        mmd_wrt_m_rst[i] = MMD(x=xp, y=yy)

    mean_mmd_st, std_mmd_st = np.mean(mmd_wrt_m_st,axis=0), np.std(mmd_wrt_m_st,axis=0)
    mean_mmd_rst, std_mmd_rst = np.mean(mmd_wrt_m_rst,axis=0), np.std(mmd_wrt_m_rst,axis=0)

    np.savez(os.path.join("mmd_data", f"metrics_{FLAGS.algorithm}_sz_{FLAGS.step_size}_m_{FLAGS.thinning_size}.npz"),
             mean_mmd_st=mean_mmd_st, std_mmd_st=std_mmd_st, mean_mmd_rst=mean_mmd_rst, std_mmd_rst=std_mmd_rst)

if __name__ == "__main__":

    app.run(main)